@@ -3,11 +3,13 @@
 """Diffusion kurtosis magnetic resonance imaging."""
 
 import argparse
+import warnings
 
 import jax
 import jax.numpy as jnp
 from jax.scipy.optimize import minimize
 import nibabel as nib
+import numba
 import numpy as np
 from sklearn.neural_network import MLPRegressor
 
@@ -15,35 +17,12 @@ from sklearn.neural_network import MLPRegressor
 jax.config.update("jax_enable_x64", True)
 jax.config.update("jax_platform_name", "cpu")
 
-SEED = 123
-
-# Parameter array elements are:
-#
-# params[..., 0] = log(S0)
-# params[..., 1] = D_xx
-# params[..., 2] = D_yy
-# params[..., 3] = D_zz
-# params[..., 4] = D_xy
-# params[..., 5] = D_xz
-# params[..., 6] = D_yz
-# params[..., 7] = W_xxxx * MD ** 2
-# params[..., 8] = W_yyyy * MD ** 2
-# params[..., 9] = W_zzzz * MD ** 2
-# params[..., 10] = W_xxxy * MD ** 2
-# params[..., 11] = W_xxxz * MD ** 2
-# params[..., 12] = W_yyyx * MD ** 2
-# params[..., 13] = W_yyyz * MD ** 2
-# params[..., 14] = W_zzzx * MD ** 2
-# params[..., 15] = W_zzzy * MD ** 2
-# params[..., 16] = W_xxyy * MD ** 2
-# params[..., 17] = W_xxzz * MD ** 2
-# params[..., 18] = W_yyzz * MD ** 2
-# params[..., 19] = W_xxyz * MD ** 2
-# params[..., 20] = W_yyxz * MD ** 2
-# params[..., 21] = W_zzxy * MD ** 2
+SEED = 123  # Hard coded for reproducible predictions
+MIN_K = -3 / 7
+MAX_K = 10
 
 
-def _design_matrix(bvals, bvecs):
+def design_matrix(bvals, bvecs):
     """Return the diffusion kurtosis imaging design matrix.
 
     Parameters
@@ -83,8 +62,8 @@ def _design_matrix(bvals, bvecs):
     return X
 
 
-def _params_to_D(params):
-    """Return the diffusion tensors corresponding to a parameter array.
+def params_to_D(params):
+    """Return diffusion tensors corresponding to a parameter array.
 
     Parameters
     ----------
@@ -108,8 +87,8 @@ def _params_to_D(params):
     return D
 
 
-def _params_to_W(params):
-    """Return the kurtosis tensors corresponding to a parameter array.
+def params_to_W(params):
+    """Return kurtosis tensors corresponding to a parameter array.
 
     Parameters
     ----------
@@ -202,16 +181,61 @@ def _params_to_W(params):
     W[..., 1, 2, 2, 0] = params[..., 21]
     W[..., 1, 2, 0, 2] = params[..., 21]
     W[..., 1, 0, 2, 2] = params[..., 21]
-    evals, _ = np.linalg.eigh(_params_to_D(params))
+    evals, _ = np.linalg.eigh(np.nan_to_num(params_to_D(params)))
     MD = np.mean(evals, axis=-1)[..., np.newaxis, np.newaxis, np.newaxis, np.newaxis]
     MD[MD == 0] = np.nan  # To avoid warnings due to division by zero
     W /= MD ** 2
     return W
 
 
+def tensors_to_params(S0, D, W):
+    """Return the parameter array corresponding to tensors.
+
+    Parameters
+    ----------
+    S0 : numpy.ndarray
+        Floating-point array.
+    D : numpy.ndarray
+        Floating-point array with shape (..., 3, 3).
+    W : numpy.ndarray
+        Floating-point array with shape (..., 3, 3, 3, 3).
+
+    Returns
+    -------
+    numpy.ndarray
+    """
+    params = np.zeros(np.asarray(S0).shape + (22,))
+    params[..., 0] = np.log(S0)
+    params[..., 1] = D[..., 0, 0]
+    params[..., 2] = D[..., 1, 1]
+    params[..., 3] = D[..., 2, 2]
+    params[..., 4] = D[..., 0, 1]
+    params[..., 5] = D[..., 0, 2]
+    params[..., 6] = D[..., 1, 2]
+    params[..., 7] = W[..., 0, 0, 0, 0]
+    params[..., 8] = W[..., 1, 1, 1, 1]
+    params[..., 9] = W[..., 2, 2, 2, 2]
+    params[..., 10] = W[..., 0, 0, 0, 1]
+    params[..., 11] = W[..., 0, 0, 0, 2]
+    params[..., 12] = W[..., 1, 1, 1, 0]
+    params[..., 13] = W[..., 1, 1, 1, 2]
+    params[..., 14] = W[..., 2, 2, 2, 0]
+    params[..., 15] = W[..., 2, 2, 2, 1]
+    params[..., 16] = W[..., 0, 0, 1, 1]
+    params[..., 17] = W[..., 0, 0, 2, 2]
+    params[..., 18] = W[..., 1, 1, 2, 2]
+    params[..., 19] = W[..., 0, 0, 1, 2]
+    params[..., 20] = W[..., 1, 1, 0, 2]
+    params[..., 21] = W[..., 2, 2, 0, 1]
+    evals, _ = np.linalg.eigh(np.nan_to_num(D))
+    MD = np.mean(evals, axis=-1)[..., np.newaxis]
+    params[..., 7::] *= MD ** 2
+    return params
+
+
 @jax.jit
 def _adc(params, vs):
-    """Compute apparent diffusion coefficients along unit vectors vs.
+    """Compute apparent diffusion coefficients along unit vectors `vs`.
 
     Parameters
     ----------
@@ -234,7 +258,7 @@ def _adc(params, vs):
     )
 
 
-def _md(params, mask=None):
+def params_to_md(params, mask=None):
     """Compute mean diffusivity.
 
     Parameters
@@ -250,13 +274,13 @@ def _md(params, mask=None):
     """
     if mask is None:
         mask = np.ones(params.shape[0:-1]).astype(bool)
-    evals, _ = np.linalg.eigh(_params_to_D(params[mask]))
+    evals, _ = np.linalg.eigh(np.nan_to_num(params_to_D(params[mask])))
     md = np.zeros(mask.shape)
     md[mask] = np.mean(evals, axis=1)
     return md
 
 
-def _ad(params, mask=None):
+def params_to_ad(params, mask=None):
     """Compute axial diffusivity.
 
     Parameters
@@ -272,13 +296,13 @@ def _ad(params, mask=None):
     """
     if mask is None:
         mask = np.ones(params.shape[0:-1]).astype(bool)
-    evals, _ = np.linalg.eigh(_params_to_D(params[mask]))
+    evals, _ = np.linalg.eigh(np.nan_to_num(params_to_D(params[mask])))
     ad = np.zeros(mask.shape)
     ad[mask] = evals[:, 2]
     return ad
 
 
-def _rd(params, mask=None):
+def params_to_rd(params, mask=None):
     """Compute radial diffusivity.
 
     Parameters
@@ -294,7 +318,7 @@ def _rd(params, mask=None):
     """
     if mask is None:
         mask = np.ones(params.shape[0:-1]).astype(bool)
-    evals, _ = np.linalg.eigh(_params_to_D(params[mask]))
+    evals, _ = np.linalg.eigh(np.nan_to_num(params_to_D(params[mask])))
     rd = np.zeros(mask.shape)
     rd[mask] = np.mean(evals[:, 0:2], axis=1)
     return rd
@@ -302,7 +326,7 @@ def _rd(params, mask=None):
 
 @jax.jit
 def _akc(params, vs):
-    """Compute apparent kurtosis coefficients along unit vectors vs.
+    """Compute apparent kurtosis coefficients along unit vectors `vs`.
 
     Parameters
     ----------
@@ -382,10 +406,10 @@ _45_dirs = np.array(
         [3.40163643e-01, 9.66360119e-02, -9.35387715e-01],
         [-1.82043727e-01, -9.27048310e-02, -9.78910566e-01],
     ]
-)
+)  # Directions along which AKC is evaluated for computing MK
 
 
-def _mk(params, mask=None):
+def params_to_mk(params, mask=None):
     """Compute mean kurtosis.
 
     Parameters
@@ -411,7 +435,7 @@ def _mk(params, mask=None):
     return mk
 
 
-def _ak(params, mask=None):
+def params_to_ak(params, mask=None):
     """Compute axial kurtosis.
 
     Parameters
@@ -431,7 +455,7 @@ def _ak(params, mask=None):
     ak = np.zeros(mask.shape)
     ak_flat = ak[mask]
     size = len(params_flat)
-    _, evecs = np.linalg.eigh(_params_to_D(params_flat))
+    _, evecs = np.linalg.eigh(np.nan_to_num(params_to_D(params_flat)))
     for i in range(size):
         ak_flat[i] = _akc(params_flat[i], evecs[np.newaxis, i, :, 2])
     ak[mask] = ak_flat
@@ -451,11 +475,12 @@ _10_dirs = np.array(
         [0.0, -9.51056516e-01, 3.09016994e-01],
         [0.0, -5.87785252e-01, 8.09016994e-01],
     ]
-)
+)  # Directions along which AKC is evaluated for computing RK
 
 
+@numba.njit
 def _vec2vec_rotmat(v, k):
-    """Compute a rotation matrix defining a rotation that aligns v with k.
+    """Compute a rotation matrix for aligning `v` with `k`.
 
     Parameters
     -----------
@@ -471,7 +496,7 @@ def _vec2vec_rotmat(v, k):
     v = v / np.linalg.norm(v)
     k = k / np.linalg.norm(k)
     axis = np.cross(v, k)
-    if np.linalg.norm(axis) < np.finfo(float).resolution:
+    if np.linalg.norm(axis) < 1e-15:
         if np.linalg.norm(v - k) > np.linalg.norm(v):
             return -np.eye(3)
         else:
@@ -482,12 +507,12 @@ def _vec2vec_rotmat(v, k):
         [[0, -axis[2], axis[1]], [axis[2], 0, -axis[0]], [-axis[1], axis[0], 0]]
     )
     R = (
-        np.eye(3) + np.sin(angle) * K + (1 - np.cos(angle)) * np.matmul(K, K)
+        np.eye(3) + np.sin(angle) * K + (1 - np.cos(angle)) * (K @ K)
     )  # Rodrigues' rotation formula
     return R
 
 
-def _rk(params, mask=None):
+def params_to_rk(params, mask=None):
     """Compute radial kurtosis.
 
     Parameters
@@ -507,7 +532,7 @@ def _rk(params, mask=None):
     rk = np.zeros(mask.shape)
     rk_flat = rk[mask]
     size = len(params_flat)
-    _, evecs = np.linalg.eigh(_params_to_D(params_flat))
+    _, evecs = np.linalg.eigh(np.nan_to_num(params_to_D(params_flat)))
     for i in range(size):
         R = _vec2vec_rotmat(np.array([1.0, 0, 0]), evecs[i, :, 2])
         vs = (R @ _10_dirs.T).T
@@ -516,6 +541,33 @@ def _rk(params, mask=None):
     return rk
 
 
+def _mtk(params, mask=None):
+    """Compute mean of the kurtosis tensor.
+
+    Parameters
+    ----------
+    params : numpy.ndarray
+        Floating-point array with shape (..., 22).
+    mask : numpy.ndarray, optional
+        Boolean array.
+
+    Returns
+    -------
+    numpy.ndarray
+    """
+    if mask is None:
+        mask = np.ones(params.shape[0:-1]).astype(bool)
+    W = params_to_W(params)
+    W_flat = W[mask]
+    mtk = np.zeros(mask.shape)
+    mtk_flat = mtk[mask]
+    size = len(W_flat)
+    for i in range(size):
+        mtk_flat[i] = np.trace(np.trace(W_flat[i])) / 5
+    mtk[mask] = mtk_flat
+    return mtk
+
+
 def _ols_fit(data, design_matrix, mask=None):
     """Estimate model parameters with ordinary least squares.
 
@@ -559,15 +611,13 @@ def _nlls_fit(data, design_matrix, mask=None):
     -------
     numpy.ndarray
     """
-
     if mask is None:
         mask = np.ones(data.shape[0:-1]).astype(bool)
 
-    data_flat = data[mask]
-
+    data_flat = jnp.asarray(data[mask])
     x0_flat = jnp.asarray(_ols_fit(data_flat, design_matrix))
     design_matrix = jnp.asarray(design_matrix)
-    data_flat = jnp.asarray(data_flat)
+    size = len(data_flat)
 
     def cost(params, design_matrix, y):
         return jnp.mean((jnp.exp(design_matrix @ params) - y) ** 2)
@@ -579,16 +629,17 @@ def _nlls_fit(data, design_matrix, mask=None):
             x0=x0_flat[i],
             args=(design_matrix, data_flat[i]),
             method="BFGS",
-            options={"line_search_maxiter": 100},
+            options={"maxiter": int(1e4), "line_search_maxiter": int(1e3)},
         )
 
-    size = len(data_flat)
     params_flat = np.zeros((size, 22))
     for i in range(size):
         results = jit_minimize(i)
-        params_flat[i] = results.x
         if not results.success:
-            print(f"Fit was not successful in voxel {i} (status = {results.status})")
+            warnings.warn(
+                f"Fit was not successful in voxel {i} (status = {results.status})"
+            )
+        params_flat[i] = results.x
 
     params = np.zeros(mask.shape + (22,))
     params[mask] = params_flat
@@ -598,7 +649,7 @@ def _nlls_fit(data, design_matrix, mask=None):
 
 @jax.jit
 def _akc_mask(W, vs, mask):
-    """Compute a binary mask based on kurtosis tensor positivity.
+    """Compute a mask based on 0 < AKC < 10 along all directions in `vs`.
 
     Parameters
     ----------
@@ -616,13 +667,13 @@ def _akc_mask(W, vs, mask):
     akc_mask = np.ones(mask.shape)
     for v in vs:
         akc_mask *= (v.T @ (v.T @ W @ v) @ v) > 0
+        akc_mask *= (v.T @ (v.T @ W @ v) @ v) < 10
     akc_mask *= mask
     return akc_mask
 
 
-def _predict(data, m, akc_mask, mask=None, **kwargs):
-    """Train a multilayer perceptron to predict `m` from `data` and return the
-    predicted map and R-squared on training data.
+def _predict(data, m, akc_mask, mask=None):
+    """Train a multilayer perceptron to predict `m` from `data`.
 
     Parameters
     ----------
@@ -630,86 +681,536 @@ def _predict(data, m, akc_mask, mask=None, **kwargs):
         Floating-point array with shape (..., number of acquisitions).
     m : numpy.ndarray
         Floating-point array.
-    akc_mask : numpy.ndarray, optional
+    akc_mask : numpy.ndarray
         Boolean array.
     mask : numpy.ndarray, optional
         Boolean array.
-    **kwargs
-        Keyword arguments passed to `sklearn.neural_network.MLPRegressor`.
 
     Returns
     -------
-    np.ndarray
-    float
+    mk_pred : np.ndarray
+    R2 : float
     """
     if mask is None:
         mask = np.ones(data.shape[0:-1]).astype(bool)
     X = data[akc_mask]
-    y = np.clip(np.nan_to_num(m[akc_mask]), -3 / 7, 10)
-    reg = MLPRegressor(random_state=SEED, **kwargs).fit(X, y)
+    y = np.clip(np.nan_to_num(m[akc_mask]), MIN_K, MAX_K)
+    reg = MLPRegressor(random_state=SEED, hidden_layer_sizes=(50, 50)).fit(X, y)
     R2 = reg.score(X, y)
     m_pred = np.zeros(mask.shape)
     m_pred[mask] = reg.predict(data[mask])
     return m_pred, R2
 
 
+def signal(params, design_matrix, mask=None):
+    """Predict signal from model parameters.
+
+    Parameters
+    ----------
+    params : numpy.ndarray
+        Floating-point array with shape (..., 22).
+    design_matrix : numpy.ndarray
+        Floating-point array with shape (number of acquisitions, 22).
+    mask : numpy.ndarray, optional
+        Boolean array.
+
+    Returns
+    -------
+    numpy.ndarray
+    """
+    if mask is None:
+        mask = np.ones(params.shape[0:-1]).astype(bool)
+    S_hat = np.exp(design_matrix @ params[..., np.newaxis])[..., 0]
+    S_hat[~mask] = 0
+    return S_hat
+
+
+def _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask=None):
+    """Calculate initial positions for the regularized NLLS fit.
+
+    Parameters
+    ----------
+    S0 : numpy.ndarray
+        Floating-point array.
+    D : numpy.ndarray
+        Floating-point array with shape (..., 3, 3).
+    mtk_pred : numpy.ndarray
+        Floating-point array.
+    ak_pred : numpy.ndarray
+        Floating-point array.
+    rk_pred : numpy.ndarray
+        Floating-point array.
+    mask : numpy.ndarray, optional
+        Boolean array.
+
+    Returns
+    -------
+    numpy.ndarray
+    """
+
+    if mask is None:
+        mask = np.ones(data.shape[0:-1]).astype(bool)
+
+    S0_flat = S0[mask]
+    D_flat = D[mask]
+    mtk_pred_flat = mtk_pred[mask]
+    ak_pred_flat = ak_pred[mask]
+    rk_pred_flat = rk_pred[mask]
+    size = len(S0_flat)
+
+    evals, evecs = np.linalg.eigh(np.nan_to_num(D_flat))
+
+    @numba.njit
+    def p(u):
+        """Return tensor P."""
+        P = np.zeros((3, 3, 3, 3))
+        for i in range(3):
+            for j in range(3):
+                for k in range(3):
+                    for l in range(3):
+                        P[i, j, k, l] = u[i] * u[j] * u[k] * u[l]
+        return P
+
+    @numba.njit
+    def d(i, j):
+        """Kronecker delta."""
+        if i == j:
+            return 1
+        else:
+            return 0
+
+    @numba.njit
+    def q(u):
+        """Return tensor Q."""
+        Q = np.zeros((3, 3, 3, 3))
+        for i in range(3):
+            for j in range(3):
+                for k in range(3):
+                    for l in range(3):
+                        Q[i, j, k, l] += (
+                            u[i] * u[j] * d(k, l)
+                            + u[i] * u[k] * d(j, l)
+                            + u[i] * u[l] * d(j, k)
+                            + u[j] * u[k] * d(i, l)
+                            + u[j] * u[l] * d(i, k)
+                            + u[k] * u[l] * d(i, j)
+                        ) / 6
+        return Q
+
+    I = np.zeros((3, 3, 3, 3))
+    for i in range(3):
+        for j in range(3):
+            for k in range(3):
+                for l in range(3):
+                    I[i, j, k, l] += (
+                        d(i, j) * d(k, l) + d(i, k) * d(j, l) + d(i, l) * d(j, k)
+                    ) / 3
+
+    x0_flat = np.zeros((size, 22))
+    for i in range(size):
+
+        u = evecs[i, :, 2]
+
+        md = np.mean(evals[i, :])
+        ad = evals[i, 2]
+        rd = np.mean(evals[i, 0:2])
+        if rd < 0.1:
+            rd = 0.1
+
+        mtk = mtk_pred_flat[i]
+        atk = ak_pred_flat[i] * ad ** 2 / md ** 2
+        rtk = rk_pred_flat[i] * rd ** 2 / md ** 2
+
+        D = rd * np.eye(3) + (ad - rd) * u[:, np.newaxis] @ u[np.newaxis, :]
+
+        W = (
+            0.5 * (10 * rtk + 5 * atk - 15 * mtk) * p(u)
+            + rtk * I
+            + 1.5 * (5 * mtk - atk - 4 * rtk) * q(u)
+        )
+
+        x0_flat[i, 0] = np.log(S0_flat[i])
+        x0_flat[i, 1] = D[0, 0]
+        x0_flat[i, 2] = D[1, 1]
+        x0_flat[i, 3] = D[2, 2]
+        x0_flat[i, 4] = D[0, 1]
+        x0_flat[i, 5] = D[0, 2]
+        x0_flat[i, 6] = D[1, 2]
+        x0_flat[i, 7] = W[0, 0, 0, 0]
+        x0_flat[i, 8] = W[1, 1, 1, 1]
+        x0_flat[i, 9] = W[2, 2, 2, 2]
+        x0_flat[i, 10] = W[0, 0, 0, 1]
+        x0_flat[i, 11] = W[0, 0, 0, 2]
+        x0_flat[i, 12] = W[1, 1, 1, 0]
+        x0_flat[i, 13] = W[1, 1, 1, 2]
+        x0_flat[i, 14] = W[2, 2, 2, 0]
+        x0_flat[i, 15] = W[2, 2, 2, 1]
+        x0_flat[i, 16] = W[0, 0, 1, 1]
+        x0_flat[i, 17] = W[0, 0, 2, 2]
+        x0_flat[i, 18] = W[1, 1, 2, 2]
+        x0_flat[i, 19] = W[0, 0, 1, 2]
+        x0_flat[i, 20] = W[1, 1, 0, 2]
+        x0_flat[i, 21] = W[2, 2, 0, 1]
+        x0_flat[i, 7::] *= md ** 2
+
+    x0 = np.zeros(mask.shape + (22,))
+    x0[mask] = x0_flat
+
+    return x0
+
+
+def _reg_nlls_fit(
+    data,
+    design_matrix,
+    x0,
+    mk_pred,
+    ak_pred,
+    rk_pred,
+    axial_dirs,
+    radial_dirs,
+    alpha=None,
+    mask=None,
+    quiet=False,
+):
+    """Estimate model parameters with regularized non-linear least squares.
+
+    Parameters
+    ----------
+    data : numpy.ndarray
+        Floating-point array with shape (..., number of acquisitions).
+    design_matrix : numpy.ndarray
+        Floating-point array with shape (number of acquisitions, 22).
+    x0 : numpy.ndarray
+        Floating-point array with shape (..., 22).
+    mk_pred : numpy.ndarray
+        Floating-point array.
+    ak_pred : numpy.ndarray
+        Floating-point array.
+    rk_pred : numpy.ndarray
+        Floating-point array.
+    axial_dirs : numpy.ndarray
+        Floating-point array with shape (..., 1, 3).
+    radial_dirs : numpy.ndarray
+        Floating-point array with shape (..., 10, 3).
+    alpha : float
+        Constant controlling regularization term magnitude.
+    mask : numpy.ndarray, optional
+        Boolean array.
+    quiet : bool
+        Whether not to print messages about computation progress.
+
+    Returns
+    -------
+    params : numpy.ndarray
+    fit_status : numpy.ndarray
+    """
+
+    if mask is None:
+        mask = np.ones(data.shape[0:-1]).astype(bool)
+
+    data_flat = jnp.asarray(data[mask])
+    design_matrix = jnp.asarray(design_matrix)
+    x0_flat = jnp.asarray(x0[mask])
+    mk_pred_flat = jnp.asarray(mk_pred[mask])
+    ak_pred_flat = jnp.asarray(ak_pred[mask])
+    rk_pred_flat = jnp.asarray(rk_pred[mask])
+    vs = jnp.asarray(_45_dirs)
+    axial_dirs_flat = jnp.asarray(axial_dirs[mask])
+    radial_dirs_flat = jnp.asarray(radial_dirs[mask])
+    size = len(data_flat)
+
+    if alpha is None:
+        mse_dki = np.median(
+            np.mean((signal(x0_flat, design_matrix) - data_flat) ** 2, axis=1)
+        )
+        mse_mk = np.median((mk_pred[mask] - params_to_mk(x0_flat)) ** 2)
+        alpha = 0.1 * mse_dki / mse_mk
+    if not quiet:
+        print(f"alpha = {alpha}")
+
+    rd_flat = jnp.asarray(params_to_rd(x0, mask)[mask])
+
+    @jax.jit
+    def cost(
+        params,
+        design_matrix,
+        y,
+        alpha,
+        vs,
+        mk_pred,
+        ak_pred,
+        rk_pred,
+        axial_dir,
+        radial_dirs,
+        rd,
+    ):
+        return (
+            jnp.mean((jnp.exp(design_matrix @ params) - y) ** 2)
+            + alpha * (jnp.mean(_akc(params, vs)) - mk_pred) ** 2
+            + alpha * (jnp.mean(_akc(params, axial_dir)) - ak_pred) ** 2
+            + alpha * (jnp.mean(_akc(params, radial_dirs)) - rk_pred) ** 2
+        )
+
+    jac = jax.jit(jax.jacfwd(cost))
+    hess = jax.jit(jax.jacfwd(jax.jacrev(cost)))
+
+    @jax.jit
+    def jit_minimize(i):
+        return minimize(
+            fun=cost,
+            x0=x0_flat[i],
+            args=(
+                design_matrix,
+                data_flat[i],
+                alpha,
+                vs,
+                mk_pred_flat[i],
+                ak_pred_flat[i],
+                rk_pred_flat[i],
+                axial_dirs_flat[i],
+                radial_dirs_flat[i],
+                rd_flat[i],
+            ),
+            method="BFGS",
+            options={
+                "maxiter": int(1e4),
+                "line_search_maxiter": int(1e3),
+                "gtol": 1e-3,
+            },
+        )
+
+    fit_status_flat = np.zeros(size)
+    params_flat = np.zeros((size, 22))
+    for i in range(size):
+        if not quiet:
+            print(f"{int(i/size*100)}%", end="\r")
+        results = jit_minimize(i)
+        params_flat[i] = results.x
+        fit_status_flat[i] = results.status
+
+    params = np.zeros(mask.shape + (22,))
+    params[mask] = params_flat
+    fit_status = np.zeros(mask.shape)
+    fit_status[mask] = fit_status_flat
+
+    return params, fit_status
+
+
+def fit(data, bvals, bvecs, mask=None, alpha=None, quiet=False):
+    """Estimate diffusion and kurtosis tensor elements.
+
+    This function does the following:
+
+        1. Remove infinities, nans, and negative values, and normalize data.
+        2. Estimate DKI model parameters using standard NLLS.
+        3. Train a multilayer perceptron to predict kurtosis measures from data
+           in voxels where the apparent kurtosis coefficient estimated by the
+           NLLS fit was > 0 and < 10 along all directions.
+        4. Estimate model parameters using regularized NLLS where the
+           regularization terms increase the objective function value when
+           MK, AK, and RK deviate from their predicted values. Axially
+           symmetric tensors with plausible magnitudes are used as initial
+           positions for the fit.
+
+    Parameters
+    ----------
+    data : numpy.ndarray
+        Floating-point array with shape (..., number of acquisitions).
+    bvals : numpy.ndarray
+        Floating-point array with shape (number of acquisitions,). Must be in
+        units of ms/μm^2!
+    bvecs : numpy.ndarray
+        Floating-point array with shape (number of acquisitions, 3).
+    mask : numpy.ndarray, optional
+        Boolean array with the same shape as `data` without the last dimension.
+    alpha : float, optional
+        Constant controlling regularization term magnitudes in the objective
+        function. 
+    quiet : bool
+        Whether not to print messages about computation progress.
+
+    Returns
+    -------
+    params : numpy.ndarray
+        Estimated model parameters:
+
+            params[..., 0] = log(S0)
+            params[..., 1] = D_xx
+            params[..., 2] = D_yy
+            params[..., 3] = D_zz
+            params[..., 4] = D_xy
+            params[..., 5] = D_xz
+            params[..., 6] = D_yz
+            params[..., 7] = W_xxxx * MD ** 2
+            params[..., 8] = W_yyyy * MD ** 2
+            params[..., 9] = W_zzzz * MD ** 2
+            params[..., 10] = W_xxxy * MD ** 2
+            params[..., 11] = W_xxxz * MD ** 2
+            params[..., 12] = W_yyyx * MD ** 2
+            params[..., 13] = W_yyyz * MD ** 2
+            params[..., 14] = W_zzzx * MD ** 2
+            params[..., 15] = W_zzzy * MD ** 2
+            params[..., 16] = W_xxyy * MD ** 2
+            params[..., 17] = W_xxzz * MD ** 2
+            params[..., 18] = W_yyzz * MD ** 2
+            params[..., 19] = W_xxyz * MD ** 2
+            params[..., 20] = W_yyxz * MD ** 2
+            params[..., 21] = W_zzxy * MD ** 2
+       
+    fit_status : numpy.ndarray
+        Codes returned by the optimizer in each voxel where it was run. 0 means
+        converged succesfully, 1 means maximum BFGS iterations reached, 3 means
+        zoom failed, 4 means saddle point reached, 5 means maximum line search
+        iterations reached, and -1 means undefined.
+    mk_pred : numpy.ndarray
+        The predicted mean kurtosis map.
+    ak_pred : numpy.ndarray
+        The predicted axial kurtosis map.
+    rk_pred : numpy.ndarray
+        The predicted radial kurtosis map.
+    """
+    if mask is None:
+        mask = np.ones(data.shape[0:-1]).astype(bool)
+
+    data[np.isinf(data)] = np.nan
+    data[np.isnan(data)] = 0
+    C = np.nanmean(data[..., np.where(bvals == np.min(bvals))])
+    data /= C
+    min_signal = np.finfo(float).resolution
+    data[data < min_signal] = min_signal
+
+    if not quiet:
+        print("Fitting DKI to data with standard NLLS")
+    X = design_matrix(bvals, bvecs)
+    nlls_params = _nlls_fit(data, X, mask)
+
+    akc_mask = _akc_mask(params_to_W(nlls_params), _45_dirs, mask).astype(bool)
+    if not quiet:
+        print("Training a MLP to predict MK")
+    mk = np.clip(params_to_mk(nlls_params, mask), MIN_K, MAX_K)
+    mk_pred, R2 = _predict(data, mk, akc_mask, mask)
+    if not quiet:
+        print(f"R^2 on training data for MK = {R2}")
+    if not quiet:
+        print("Training a MLP to predict MTK")
+    mtk = np.clip(_mtk(nlls_params, mask), MIN_K, MAX_K)
+    mtk_pred, R2 = _predict(data, mtk, akc_mask, mask)
+    if not quiet:
+        print(f"R^2 on training data for MTK = {R2}")
+    if not quiet:
+        print("Training a MLP to predict AK")
+    ak = np.clip(params_to_ak(nlls_params, mask), MIN_K, MAX_K)
+    ak_pred, R2 = _predict(data, ak, akc_mask, mask)
+    if not quiet:
+        print(f"R^2 on training data for AK = {R2}")
+    if not quiet:
+        print("Training a MLP to predict RK")
+    rk = np.clip(params_to_rk(nlls_params, mask), MIN_K, MAX_K)
+    rk_pred, R2 = _predict(data, rk, akc_mask, mask)
+    if not quiet:
+        print(f"R^2 on training data for RK = {R2}")
+
+    if not quiet:
+        print("Calculating initial positions")
+    axial_dirs = np.zeros(mask.shape + (1, 3))
+    axial_dirs_flat = axial_dirs[mask]
+    radial_dirs = np.zeros(mask.shape + (10, 3))
+    radial_dirs_flat = radial_dirs[mask]
+    size = len(radial_dirs_flat)
+    _, evecs = np.linalg.eigh(np.nan_to_num(params_to_D(nlls_params)))
+    evecs_flat = evecs[mask]
+    for i in range(size):
+        axial_dirs_flat[i] = evecs_flat[i, :, 2].T
+        R = _vec2vec_rotmat(np.array([1.0, 0, 0]), evecs_flat[i, :, 2])
+        radial_dirs_flat[i] = (R @ _10_dirs.T).T
+    axial_dirs[mask] = axial_dirs_flat
+    radial_dirs[mask] = radial_dirs_flat
+    S0 = np.exp(nlls_params[..., 0])
+    D = params_to_D(nlls_params)
+    x0 = _calculate_x0(S0, D, mtk_pred, ak_pred, rk_pred, mask)
+
+    if not quiet:
+        print("Fitting DKI to data with regularized NLLS")
+    params, fit_status = _reg_nlls_fit(
+        data,
+        X,
+        x0,
+        mk_pred,
+        ak_pred,
+        rk_pred,
+        axial_dirs,
+        radial_dirs,
+        alpha,
+        mask,
+        quiet,
+    )
+
+    return params, fit_status, mk_pred, ak_pred, rk_pred
+
+
 if __name__ == "__main__":
 
     # Parse arguments
 
     parser = argparse.ArgumentParser(
         description=(
-            "Estimate diffusion and kurtosis tensors, and compute parameter maps."
+            "Estimate diffusion and kurtosis tensors, and compute parameter maps"
         )
     )
     parser.add_argument(
-        "data", help="path of a NIfTI file with diffusion-weighted data.",
+        "data", help="path of a NIfTI file with diffusion-weighted data",
     )
     parser.add_argument(
-        "bvals", help="path of a text file with b-values in units of s/mm^2.",
+        "bvals", help="path of a text file with b-values in units of s/mm^2",
     )
     parser.add_argument(
-        "bvecs", help="path of a text file with b-vectors.",
+        "bvecs", help="path of a text file with b-vectors",
     )
     parser.add_argument(
         "-mask",
         help=(
             "path of a NIfTI file with a binary mask definining the voxels where to "
-            "estimate the tensors."
+            "estimate the tensors"
         ),
     )
     parser.add_argument(
-        "-md", help="path of a file in which to save the mean diffusivity map."
+        "-md", help="path of a file in which to save the mean diffusivity map"
     )
     parser.add_argument(
-        "-ad", help="path of a file in which to save the axial diffusivity map.",
+        "-ad", help="path of a file in which to save the axial diffusivity map",
     )
     parser.add_argument(
-        "-rd", help="path of a file in which to save the radial diffusivity map.",
+        "-rd", help="path of a file in which to save the radial diffusivity map",
     )
     parser.add_argument(
-        "-mk", help="path of a file in which to save the mean kurtosis map."
+        "-mk", help="path of a file in which to save the mean kurtosis map"
     )
     parser.add_argument(
-        "-ak", help="path of a file in which to save the axial kurtosis map.",
+        "-ak", help="path of a file in which to save the axial kurtosis map",
     )
     parser.add_argument(
-        "-rk", help="path of a file in which to save the radial kurtosis map.",
+        "-rk", help="path of a file in which to save the radial kurtosis map",
     )
     parser.add_argument(
-        "-s0", help="path of a file in which to save the estimated signal at b=0.",
+        "-s0", help="path of a file in which to save the estimated signal at b=0",
     )
     parser.add_argument(
         "-mk_pred",
-        help="path of a file in which to save the predicted mean kurtosis map.",
+        help="path of a file in which to save the predicted mean kurtosis map",
     )
     parser.add_argument(
         "-ak_pred",
-        help="path of a file in which to save the predicted axial kurtosis map.",
+        help="path of a file in which to save the predicted axial kurtosis map",
     )
     parser.add_argument(
         "-rk_pred",
-        help="path of a file in which to save the predicted radial kurtosis map.",
+        help="path of a file in which to save the predicted radial kurtosis map",
+    )
+    parser.add_argument(
+        "-fit_status", help="path of a file in which to save the fit status map",
+    )
+    parser.add_argument(
+        "-params", help="path of a file in which to save the parameters",
     )
     args = parser.parse_args()
 
@@ -727,54 +1228,26 @@ if __name__ == "__main__":
     else:
         mask = None
 
-    # Clean, scale, and clip data
+    # Fit model to data with regularized NLLS
 
-    data[np.isinf(data)] = np.nan
-    data[np.isnan(data)] = 0
-    C = np.nanmean(data[..., np.where(bvals == np.min(bvals))])
-    data /= C
-    min_signal = np.finfo(float).resolution
-    data[data < min_signal] = min_signal
-
-    # Fit model to data
-
-    print("Fitting DKI to data with standard NLLS")
-    X = _design_matrix(bvals, bvecs)
-    params = _nlls_fit(data, X, mask)
-
-    if args.mk_pred:
-        print("Training a MLP to predict MK")
-        mk = np.clip(_mk(params, mask), -3 / 7, 10)
-        akc_mask = _akc_mask(_params_to_W(params), _45_dirs, mask).astype(bool)
-        mk_pred, R2 = _predict(data, mk, akc_mask, mask, hidden_layer_sizes=(50, 50))
-        print(f"R^2 on training data = {R2}")
-    if args.ak_pred:
-        print("Training a MLP to predict AK")
-        ak = np.clip(_ak(params, mask), -3 / 7, 10)
-        akc_mask = _akc_mask(_params_to_W(params), _45_dirs, mask).astype(bool)
-        ak_pred, R2 = _predict(data, ak, akc_mask, mask, hidden_layer_sizes=(50, 50))
-        print(f"R^2 on training data = {R2}")
-    if args.rk_pred:
-        print("Training a MLP to predict RK")
-        rk = np.clip(_rk(params, mask), -3 / 7, 10)
-        akc_mask = _akc_mask(_params_to_W(params), _45_dirs, mask).astype(bool)
-        rk_pred, R2 = _predict(data, rk, akc_mask, mask, hidden_layer_sizes=(50, 50))
-        print(f"R^2 on training data = {R2}")
+    params, fit_status, mk_pred, ak_pred, rk_pred = fit(data, bvals, bvecs, mask)
 
     # Save results
 
+    if args.params:
+        nib.save(nib.Nifti1Image(params, affine), args.params)
     if args.md:
-        nib.save(nib.Nifti1Image(_md(params, mask) * 1e-3, affine), args.md)
+        nib.save(nib.Nifti1Image(params_to_md(params, mask) * 1e-3, affine), args.md)
     if args.ad:
-        nib.save(nib.Nifti1Image(_ad(params, mask) * 1e-3, affine), args.ad)
+        nib.save(nib.Nifti1Image(params_to_ad(params, mask) * 1e-3, affine), args.ad)
     if args.rd:
-        nib.save(nib.Nifti1Image(_rd(params, mask) * 1e-3, affine), args.rd)
+        nib.save(nib.Nifti1Image(params_to_rd(params, mask) * 1e-3, affine), args.rd)
     if args.mk:
-        nib.save(nib.Nifti1Image(_mk(params, mask), affine), args.mk)
+        nib.save(nib.Nifti1Image(params_to_mk(params, mask), affine), args.mk)
     if args.ak:
-        nib.save(nib.Nifti1Image(_ak(params, mask), affine), args.ak)
+        nib.save(nib.Nifti1Image(params_to_ak(params, mask), affine), args.ak)
     if args.rk:
-        nib.save(nib.Nifti1Image(_rk(params, mask), affine), args.rk)
+        nib.save(nib.Nifti1Image(params_to_rk(params, mask), affine), args.rk)
     if args.s0:
         nib.save(nib.Nifti1Image(np.exp(params[..., 0] + np.log(C)), affine), args.s0)
     if args.mk_pred:
@@ -783,3 +1256,6 @@ if __name__ == "__main__":
         nib.save(nib.Nifti1Image(ak_pred, affine), args.ak_pred)
     if args.rk_pred:
         nib.save(nib.Nifti1Image(rk_pred, affine), args.rk_pred)
+    if args.fit_status:
+        nib.save(nib.Nifti1Image(fit_status, affine), args.fit_status)
+
